/* QCharSelect is a fork of KCharSelect from KDE 3.5.10, ported to pure Qt
 *
 * QCharSelect is Copyright (c) 2013 D. Michael McIntyre
 *
 * In KDE, KCharSelect the application was built around KCharSelect the library
 * widget.  QCharSelect extracts both components from KDE, moving the
 * application's core functionality into QCharSelectDialog, and moving the
 * library widget's core functionality into QCharSelectWidget, both of which
 * compile and link together to produce the qcharselect binary.
 *
 * The KCharSelect application was originally:
 *
 *     Copyright (C) 1999 Reginald Stadlbauer <reggie@kde.org>
 *
 * It was released under the GPL.
 *
 * The KCharSelect library widget was originally:
 *
 *     Copyright (C) 1999 Reginald Stadlbauer <reggie@kde.org>
 *
 * It was relesaed under the LGPL.  Under the terms of the LGPL, I am exercising
 * the option to release this derivative work under the GPL.  As such, the LGPL
 * notice has been removed from any source files where it was originally
 * present.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.  See the file COPYING included with this distribution for more
 * information.
*/

#include "QCharSelectWidget.h"

#include <Q3GridView>
#include <Q3ScrollView>

#include <QBrush>
#include <QColor>
#include <QEvent>
#include <QFont>
#include <QFontDatabase>
#include <QWidget>
#include <QHBoxLayout>
#include <QLabel>
#include <QPainter>
#include <QPen>
#include <QRegExp>
#include <QStyle>
#include <QToolTip>
#include <QValidator>
#include <QApplication>
#include <QDebug>
#include <QDialog>
#include <QLineEdit>
#include <QMouseEvent>
#include <QResizeEvent>
#include <QLinkedList>
#include <QFrame>
#include <QGroupBox>

#include <iostream>

class QCharSelectWidget::QCharSelectWidgetPrivate
{
public:
    QLineEdit *unicodeLine;
};

QFontDatabase *QCharSelectWidget::fontDataBase = 0;

void QCharSelectWidget::cleanupFontDatabase()
{
    delete fontDataBase;
    fontDataBase = 0;
}


/******************************************************************/
/* Class: QCharSelectWidgetTable                                  */
/******************************************************************/

//==================================================================
QCharSelectWidgetTable::QCharSelectWidgetTable(const QString &_font,
                                               const QChar &_chr,
                                               int _tableNum)
    : Q3GridView(),
        vFont(_font),
        vChr(_chr),
        vTableNum(_tableNum),
        vPos(0, 0),
        focusItem(_chr),
        focusPos(0, 0),
        d(0)
{
    std::cerr << "QCharSelectWidgetTable - ctor" << std::endl;

    setStyleSheet("background: white");
//    setBackgroundColor(colorGroup().base());

    setCellWidth(20);
    setCellHeight(25);

    setNumCols(32);
    setNumRows(8);

    repaintContents(false);
    
    setToolTips();

    //!!!
    //setFocusPolicy(QWidget::StrongFocus);
    setBackgroundMode(Qt::NoBackground);
}

//==================================================================
void QCharSelectWidgetTable::setFont(const QString &_font)
{
    vFont = _font;
    repaintContents(false);

    setToolTips();
}

//==================================================================
void QCharSelectWidgetTable::setChar(const QChar &_chr)
{
    vChr = _chr;
    repaintContents(false);
}

//==================================================================
void QCharSelectWidgetTable::setTableNum(int _tableNum)
{
    focusItem = QChar(_tableNum * 256);

    vTableNum = _tableNum;
    repaintContents(false);

    setToolTips();
}

//==================================================================
QSize QCharSelectWidgetTable::sizeHint() const
{
    int w = cellWidth();
    int h = cellHeight();

    w *= numCols();
    h *= numRows();

    return QSize(w, h);
}

//==================================================================
void QCharSelectWidgetTable::resizeEvent(QResizeEvent * e)
{
    const int new_w   = (e->size().width()  - 2*(margin()+frameWidth())) / numCols();
    const int new_h   = (e->size().height() - 2*(margin()+frameWidth())) / numRows();

    if(new_w !=  cellWidth())
        setCellWidth(new_w);
    if(new_h !=  cellHeight())
        setCellHeight(new_h);

    setToolTips();
}

//==================================================================
void QCharSelectWidgetTable::paintCell(class QPainter* p, int row, int col)
{
    const int w = cellWidth();
    const int h = cellHeight();
    const int x2 = w - 1;
    const int y2 = h - 1;

    //if(row == 0 && col == 0) {
    //    printf("Repaint %d\n", temp++);
    //    fflush(stdout);
    //    }

    QFont font = QFont(vFont);
    font.setPixelSize(int(.7 * h));

    unsigned short c = vTableNum * 256;
    c += row * numCols();
    c += col;

    if (c == vChr.unicode()) {
    p->setBrush(QBrush(colorGroup().highlight()));
    p->setPen(Qt::NoPen);
    p->drawRect(0, 0, w, h);
    p->setPen(colorGroup().highlightedText());
    vPos = QPoint(col, row);
    } else {
    QFontMetrics fm = QFontMetrics(font);
    if(fm.inFont(c))
        p->setBrush(QBrush(colorGroup().base()));
    else
        p->setBrush(QBrush(colorGroup().button()));
    p->setPen(Qt::NoPen);
    p->drawRect(0, 0, w, h);
    p->setPen(colorGroup().text());
    }

    if (c == focusItem.unicode() && hasFocus()) {
//!!!
//        style().drawPrimitive(QStyle::PE_FocusRect, p, QRect(2, 2, w - 4, h - 4), 
//                              colorGroup());
        focusPos = QPoint(col, row);
    }

    p->setFont(font);

    p->drawText(0, 0, x2, y2, Qt::AlignHCenter | Qt::AlignVCenter, QString(QChar(c)));

    p->setPen(colorGroup().text());
    p->drawLine(x2, 0, x2, y2);
    p->drawLine(0, y2, x2, y2);

    if (row == 0)
    p->drawLine(0, 0, x2, 0);
    if (col == 0)
    p->drawLine(0, 0, 0, y2);
}

//==================================================================
void QCharSelectWidgetTable::mouseMoveEvent(QMouseEvent *e)
{
    const int row = rowAt(e->y());
    const int col = columnAt(e->x());
    if (row >= 0 && row < numRows() && col >= 0 && col < numCols()) {
    const QPoint oldPos = vPos;

    vPos.setX(col);
    vPos.setY(row);

    vChr = QChar(vTableNum * 256 + numCols() * vPos.y() + vPos.x());

    const QPoint oldFocus = focusPos;

    focusPos = vPos;
    focusItem = vChr;

    repaintCell(oldFocus.y(), oldFocus.x(), true);
    repaintCell(oldPos.y(), oldPos.x(), true);
    repaintCell(vPos.y(), vPos.x(), true);

    emit highlighted(vChr);
    emit highlighted();

    emit focusItemChanged(focusItem);
    emit focusItemChanged();
    }
}

//==================================================================
void QCharSelectWidgetTable::keyPressEvent(QKeyEvent *e)
{
    switch (e->key()) {
    case Qt::Key_Left:
            gotoLeft();
            break;
    case Qt::Key_Right:
            gotoRight();
            break;
    case Qt::Key_Up:
            gotoUp();
            break;
    case Qt::Key_Down:
            gotoDown();
            break;
    case Qt::Key_Next:
            emit tableDown();
            break;
    case Qt::Key_Prior:
            emit tableUp();
            break;
    case Qt::Key_Space:
            emit activated(' ');
            emit activated();
            emit highlighted(' ');
            emit highlighted();
                break;
    case Qt::Key_Enter:
    case Qt::Key_Return: {
         const QPoint oldPos = vPos;

         vPos = focusPos;
         vChr = focusItem;

         repaintCell(oldPos.y(), oldPos.x(), true);
         repaintCell(vPos.y(), vPos.x(), true);

         emit activated(vChr);
         emit activated();
         emit highlighted(vChr);
         emit highlighted();
         }
         break;
    }
}

//==================================================================
void QCharSelectWidgetTable::gotoLeft()
{
    if (focusPos.x() > 0) {
    const QPoint oldPos = focusPos;

    focusPos.setX(focusPos.x() - 1);

    focusItem = QChar(vTableNum * 256 + numCols() * focusPos.y() + focusPos.x());

    repaintCell(oldPos.y(), oldPos.x(), true);
    repaintCell(focusPos.y(), focusPos.x(), true);

    emit focusItemChanged(vChr);
    emit focusItemChanged();
    }
}

//==================================================================
void QCharSelectWidgetTable::gotoRight()
{
    if (focusPos.x() < numCols()-1) {
    const QPoint oldPos = focusPos;

    focusPos.setX(focusPos.x() + 1);

    focusItem = QChar(vTableNum * 256 + numCols() * focusPos.y() + focusPos.x());

    repaintCell(oldPos.y(), oldPos.x(), true);
    repaintCell(focusPos.y(), focusPos.x(), true);

    emit focusItemChanged(vChr);
    emit focusItemChanged();
    }
}

//==================================================================
void QCharSelectWidgetTable::gotoUp()
{
    if (focusPos.y() > 0) {
    const QPoint oldPos = focusPos;

    focusPos.setY(focusPos.y() - 1);

    focusItem = QChar(vTableNum * 256 + numCols() * focusPos.y() + focusPos.x());

    repaintCell(oldPos.y(), oldPos.x(), true);
    repaintCell(focusPos.y(), focusPos.x(), true);

    emit focusItemChanged(vChr);
    emit focusItemChanged();
    }
}

//==================================================================
void QCharSelectWidgetTable::gotoDown()
{
    if (focusPos.y() < numRows()-1) {
    const QPoint oldPos = focusPos;

    focusPos.setY(focusPos.y() + 1);

    focusItem = QChar(vTableNum * 256 + numCols() * focusPos.y() + focusPos.x());

    repaintCell(oldPos.y(), oldPos.x(), true);
    repaintCell(focusPos.y(), focusPos.x(), true);

    emit focusItemChanged(vChr);
    emit focusItemChanged();
    }
}

//==================================================================
void QCharSelectWidgetTable::setToolTips()
{
    const int rowCount = numRows();
    const int colCount = numCols();
    for (int i = 0 ; i < rowCount; ++i) {
        for (int j = 0; j < colCount; ++j)  {
            const QRect r(cellWidth()*j, cellHeight()*i, cellWidth(), cellHeight());
            setToolTip("");
            const ushort uni = vTableNum * 256 + numCols()*i + j;
            QString s;
            s.sprintf("%04X", uint(uni));
            QString character; // Character (which sometimes need to be escaped)
            switch (uni)
            {
                    case 0x3c: character = "&lt;"; break;
                    case 0x3e: character = "&gt;"; break;
                    case 0x26: character = "&amp;"; break;
                    case 0x27: character = "&apos;"; break;
                    case 0x22: character = "&quot;"; break;
                    default: character = QChar(uni); break;
            }
            setToolTip(tr("<qt><font size=\"+4\" face=\"%1\">%2</font><br>Unicode code point: U+%3<br>(In decimal: %4)<br>(Character: %5)</qt>").arg(vFont).arg(character).arg(s).arg(uni).arg(character));
        }
    }
}

/******************************************************************/
/* Class: QCharSelectWidget                                       */
/******************************************************************/

//==================================================================
QCharSelectWidget::QCharSelectWidget(const QString &_font,
                                     const QChar &_chr,
                                     int _tableNum)
  : QWidget(),
      d(new QCharSelectWidgetPrivate)
{
    std::cerr << "QCharSelectWidget - ctor" << std::endl;

   // setStyleSheet("background: #FF00EE; color: yellow");

    QVBoxLayout *vboxLayout = new QVBoxLayout;
    setLayout(vboxLayout);
    //!!!
    //vboxLayout->setSpacing(QDialog::spacingHint());
    
    QGroupBox *bar = new QGroupBox;
    QHBoxLayout *barLayout = new QHBoxLayout;
    bar->setLayout(barLayout);
    vboxLayout->addWidget(bar);
    //!!!
    //barLayout->setSpacing(QDialog::spacingHint());

    QLabel *lFont = new QLabel(tr("Font:"));
    barLayout->addWidget(lFont);

    lFont->resize(lFont->sizeHint());
    lFont->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
    lFont->setMaximumWidth(lFont->sizeHint().width());

    fontCombo = new QComboBox();
    fontCombo->setEditable(true);
    barLayout->addWidget(fontCombo);
    fillFontCombo();
    fontCombo->resize(fontCombo->sizeHint());

    connect(fontCombo, SIGNAL(activated(const QString &)), this, SLOT(fontSelected(const QString &)));

    QLabel* const lTable = new QLabel(tr("Table:"));
    barLayout->addWidget(lTable);
    lTable->resize(lTable->sizeHint());
    lTable->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
    lTable->setMaximumWidth(lTable->sizeHint().width());

    tableSpinBox = new QSpinBox();
    tableSpinBox->setMinimum(0);
    tableSpinBox->setMaximum(255);
    tableSpinBox->setSingleStep(1);
    barLayout->addWidget(tableSpinBox);
    tableSpinBox->resize(tableSpinBox->sizeHint());

    connect(tableSpinBox, SIGNAL(valueChanged(int)), this, SLOT(tableChanged(int)));

    QLabel* const lUnicode = new QLabel(tr("&Unicode code point:"));
    barLayout->addWidget(lUnicode);
    lUnicode->resize(lUnicode->sizeHint());
    lUnicode->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
    lUnicode->setMaximumWidth(lUnicode->sizeHint().width());

    const QRegExp rx("[a-fA-F0-9]{1,4}");
    QValidator* const validator = new QRegExpValidator(rx, this);

    d->unicodeLine = new QLineEdit();
    barLayout->addWidget(d->unicodeLine);
    d->unicodeLine->setValidator(validator);
    lUnicode->setBuddy(d->unicodeLine);
    d->unicodeLine->resize(d->unicodeLine->sizeHint());
    slotUpdateUnicode(_chr);

    connect(d->unicodeLine, SIGNAL(returnPressed()), this, SLOT(slotUnicodeEntered()));

    QFont currentFont;

    charTable = new QCharSelectWidgetTable(_font.isEmpty() ? currentFont.family() : _font, _chr, _tableNum);
    vboxLayout->addWidget(charTable);
    const QSize sz(charTable->contentsWidth()  +  4 ,
                    charTable->contentsHeight() +  4);
    charTable->resize(sz);
    //charTable->setMaximumSize(sz);
    charTable->setMinimumSize(sz);
    charTable->setHScrollBarMode(Q3ScrollView::AlwaysOff);
    charTable->setVScrollBarMode(Q3ScrollView::AlwaysOff);

    setFont(_font.isEmpty() ? currentFont.family() : _font);
    setTableNum(_tableNum);

    connect(charTable, SIGNAL(highlighted(const QChar &)), this, SLOT(slotUpdateUnicode(const QChar &)));
    connect(charTable, SIGNAL(highlighted(const QChar &)), this, SLOT(charHighlighted(const QChar &)));
    connect(charTable, SIGNAL(highlighted()), this, SLOT(charHighlighted()));
    connect(charTable, SIGNAL(activated(const QChar &)), this, SLOT(charActivated(const QChar &)));
    connect(charTable, SIGNAL(activated()), this, SLOT(charActivated()));
    connect(charTable, SIGNAL(focusItemChanged(const QChar &)),
                              this, SLOT(charFocusItemChanged(const QChar &)));
    connect(charTable, SIGNAL(focusItemChanged()), this, SLOT(charFocusItemChanged()));
    connect(charTable, SIGNAL(tableUp()), this, SLOT(charTableUp()));
    connect(charTable, SIGNAL(tableDown()), this, SLOT(charTableDown()));

    connect(charTable, SIGNAL(doubleClicked()),this,SLOT(slotDoubleClicked()));

    //!!!
    //setFocusPolicy(QWidget::StrongFocus);
    setFocusProxy(charTable);/* */
}

QCharSelectWidget::~QCharSelectWidget()
{
    delete d;
}

//==================================================================
QSize QCharSelectWidget::sizeHint() const
{
    // we *are* the "QVBox" now
    return QWidget::sizeHint();
//    return QVBox::sizeHint();
}

//==================================================================
void QCharSelectWidget::setFont(const QString &_font)
{
    QList<QString>::iterator it = fontList.find(_font);
    if (it != fontList.end()) {
        QList<QString>::iterator it2 = fontList.begin();
        int pos = 0;
        for (; it != it2; ++it2, ++pos);
        fontCombo->setCurrentIndex(pos);
        charTable->setFont(_font);
    }
//    else
        //!!!
        //kdWarning() << "Can't find Font: " << _font << endl;
}

//==================================================================
void QCharSelectWidget::setChar(const QChar &_chr)
{
    charTable->setChar(_chr);
    slotUpdateUnicode(_chr);
}

//==================================================================
void QCharSelectWidget::setTableNum(int _tableNum)
{
    tableSpinBox->setValue(_tableNum);
    charTable->setTableNum(_tableNum);
}

//==================================================================
void QCharSelectWidget::fillFontCombo()
{
    if (!fontDataBase) {
        fontDataBase = new QFontDatabase();
        qAddPostRoutine(cleanupFontDatabase);
    }
    fontList=fontDataBase->families();
    fontCombo->addItems(fontList);
}

//==================================================================
void QCharSelectWidget::fontSelected(const QString &_font)
{
    charTable->setFont(_font);
    emit fontChanged(_font);
}

//==================================================================
void QCharSelectWidget::tableChanged(int _value)
{
    charTable->setTableNum(_value);
}

//==================================================================
void QCharSelectWidget::slotUnicodeEntered()
{
    const QString s = d->unicodeLine->text();
    if (s.isEmpty())
        return;
    
    bool ok;
    const int uc = s.toInt(&ok, 16);
    if (!ok)
        return;
    
    const int table = uc / 256;
    charTable->setTableNum(table);
    tableSpinBox->setValue(table);
    const QChar ch(uc);
    charTable->setChar(ch);
    charActivated(ch);
}

void QCharSelectWidget::slotUpdateUnicode(const QChar &c)
{
    const int uc = c.unicode();
    QString s;
    s.sprintf("%04X", uc);
    d->unicodeLine->setText(s);
}

void QCharSelectWidgetTable::virtual_hook(int, void*)
{ /*BASE::virtual_hook(id, data);*/ }

void QCharSelectWidget::virtual_hook(int, void*)
{ /*BASE::virtual_hook(id, data);*/ }

#include "QCharSelectWidget.moc"
